"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) if (e.indexOf(p[i]) < 0)
            t[p[i]] = s[p[i]];
    return t;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const events_1 = require("events");
const graphql_1 = require("graphql");
const chalk_1 = __importDefault(require("chalk"));
const path_1 = require("path");
const fs_extra_1 = require("fs-extra");
const chokidar_1 = require("chokidar");
const glob = __importStar(require("glob"));
const graphql_config_1 = require("graphql-config");
const ast_1 = require("graphql-tool-utilities/ast");
const config_1 = require("graphql-tool-utilities/config");
const print_1 = require("./print");
const types_1 = require("./types");
exports.EnumFormat = types_1.EnumFormat;
class Builder extends events_1.EventEmitter {
    constructor(_a) {
        var { cwd } = _a, options = __rest(_a, ["cwd"]);
        super();
        // projectName -> {filePath -> document}
        // NOTE: projectName can be undefined for nameless graphql-config projects
        this.documentMapByProject = new Map();
        this.watchers = [];
        this.options = options;
        this.config = graphql_config_1.getGraphQLConfig(cwd ? path_1.resolve(cwd) : undefined);
    }
    once(event, handler) {
        return super.once(event, handler);
    }
    on(event, handler) {
        return super.on(event, handler);
    }
    emit(event, ...args) {
        return super.emit(event, ...args);
    }
    run({ watch: watchGlobs = false } = {}) {
        return __awaiter(this, void 0, void 0, function* () {
            const schemaPaths = config_1.getGraphQLSchemaPaths(this.config);
            if (watchGlobs) {
                this.watchers.push(...this.setupDocumentWatchers().concat(this.setupSchemaWatcher()));
                // wait for all watchers to be ready
                yield Promise.all(this.watchers.map((watcher) => new Promise((resolve) => watcher.on('ready', () => resolve()))));
            }
            try {
                this.emit('start:schema');
                schemaPaths.forEach((schemaPath) => this.generateSchemaTypes(schemaPath));
                this.emit('end:schema');
            }
            catch (error) {
                this.emit('error', error);
                return;
            }
            try {
                config_1.getGraphQLProjects(this.config).forEach((projectConfig) => this.updateDocumentsForProject(projectConfig));
            }
            catch (error) {
                this.emit('error', error);
                return;
            }
            this.generateDocumentTypes();
        });
    }
    stop() {
        this.watchers.forEach((watcher) => {
            watcher.close();
        });
        this.watchers.length = 0;
    }
    setupDocumentWatchers() {
        const update = (filePath, projectConfig) => {
            try {
                this.updateDocumentForFile(filePath, projectConfig);
            }
            catch (error) {
                this.emit('error', error);
                return;
            }
            this.generateDocumentTypes();
        };
        return config_1.getGraphQLProjects(this.config).map((projectConfig) => {
            return chokidar_1.watch(projectConfig.includes.map((include) => config_1.getGraphQLFilePath(this.config, include)), {
                ignored: projectConfig.excludes.map((exclude) => config_1.getGraphQLFilePath(this.config, exclude)),
                ignoreInitial: true,
            })
                .on('add', (filePath) => update(filePath, projectConfig))
                .on('change', (filePath) => update(filePath, projectConfig))
                .on('unlink', (filePath) => {
                const documents = this.documentMapByProject.get(projectConfig.projectName);
                if (documents) {
                    documents.delete(filePath);
                }
                this.generateDocumentTypes();
            });
        });
    }
    setupSchemaWatcher() {
        const update = (schemaPath) => {
            try {
                this.emit('start:schema');
                this.generateSchemaTypes(schemaPath);
                this.emit('end:schema');
                this.generateDocumentTypes();
            }
            catch (error) {
                // intentional noop
            }
        };
        return chokidar_1.watch(config_1.getGraphQLSchemaPaths(this.config), { ignoreInitial: true }).on('change', update);
    }
    generateSchemaTypes(schemaPath) {
        const projectConfig = config_1.getGraphQLProjectForSchemaPath(this.config, schemaPath);
        const schemaTypesPath = getSchemaTypesPath(projectConfig, this.options);
        const definition = print_1.printSchema(projectConfig.getSchema(), this.options);
        fs_extra_1.mkdirpSync(path_1.dirname(schemaTypesPath));
        fs_extra_1.writeFileSync(schemaTypesPath, definition);
        this.emit('build:schema', {
            schemaPath,
            schemaTypesPath,
        });
    }
    generateDocumentTypes() {
        this.emit('start:docs');
        getDuplicateOperations(this.documentMapByProject).forEach(({ projectName, duplicates }) => {
            if (duplicates.length) {
                duplicates.forEach(({ operationName, filePaths }) => {
                    const message = `GraphQL operations must have a unique name. The operation ${chalk_1.default.bold(operationName)} is declared in:\n ${filePaths.sort().join('\n ')}${projectName ? ` (${chalk_1.default.bold(projectName)})` : ''}`;
                    this.emit('error', new Error(message));
                });
            }
        });
        for (const [projectName, documents,] of this.documentMapByProject.entries()) {
            this.generateDocumentTypesForProject(this.config.getProjectConfig(projectName), documents);
        }
        this.emit('end:docs');
    }
    generateDocumentTypesForProject(projectConfig, documents) {
        let ast;
        try {
            ast = ast_1.compile(projectConfig.getSchema(), graphql_1.concatAST(Array.from(documents.values())));
        }
        catch (error) {
            this.emit('error', error);
            return;
        }
        const fileMap = groupOperationsAndFragmentsByFile(ast);
        try {
            for (const file of fileMap.values()) {
                this.writeDocumentFile(file, ast, projectConfig);
            }
        }
        catch (error) {
            // intentional noop
        }
    }
    writeDocumentFile(file, ast, project) {
        const definitionPath = `${file.path}.d.ts`;
        const definition = this.getDocumentDefinition(file, ast, project);
        fs_extra_1.writeFileSync(definitionPath, definition);
        const result = {
            documentPath: file.path,
            definitionPath,
            operation: file.operation,
            fragments: file.fragments,
        };
        this.emit('build:docs', result);
    }
    getDocumentDefinition(file, ast, project) {
        try {
            return print_1.printDocument(file, ast, Object.assign({}, this.options, { schemaTypesPath: getSchemaTypesPath(project, this.options) }));
        }
        catch ({ message }) {
            const error = new Error(`Error in ${file.path}: ${message[0].toLowerCase()}${message.slice(1)}`);
            this.emit('error', error);
            throw error;
        }
    }
    updateDocumentsForProject(projectConfig) {
        return Array.from(projectConfig.includes
            .flatMap((include) => glob.sync(config_1.getGraphQLFilePath(this.config, include), {
            ignore: projectConfig.excludes.map((exclude) => config_1.getGraphQLFilePath(this.config, exclude)),
        }))
            .reduce((filePaths, filePath) => {
            return filePaths.add(filePath);
        }, new Set())
            .values()).map((filePath) => this.updateDocumentForFile(filePath, projectConfig));
    }
    updateDocumentForFile(filePath, project) {
        let documents = this.documentMapByProject.get(project.projectName);
        if (!documents) {
            documents = new Map();
            this.documentMapByProject.set(project.projectName, documents);
        }
        const contents = fs_extra_1.readFileSync(filePath, 'utf8');
        if (contents.trim().length === 0) {
            return undefined;
        }
        const document = graphql_1.parse(new graphql_1.Source(contents, filePath));
        documents.set(filePath, document);
        return document;
    }
}
exports.Builder = Builder;
function getSchemaTypesPath(project, options) {
    if (typeof project.extensions.schemaTypesPath === 'string') {
        return config_1.getGraphQLFilePath(project, project.extensions.schemaTypesPath);
    }
    return config_1.getGraphQLFilePath(project, path_1.join(options.schemaTypesPath, `${project.projectName ? `${project.projectName}-` : ''}types.ts`));
}
function groupOperationsAndFragmentsByFile({ operations, fragments }) {
    return Object.values(operations)
        .concat(Object.values(fragments))
        .reduce((map, item) => {
        if (!item.filePath) {
            return map;
        }
        let file = map.get(item.filePath);
        if (!file) {
            file = {
                path: item.filePath,
                operation: undefined,
                fragments: [],
            };
            map.set(item.filePath, file);
        }
        if (ast_1.isOperation(item)) {
            file.operation = item;
        }
        else {
            file.fragments.push(item);
        }
        return map;
    }, new Map());
}
function getDuplicateOperations(documentsMapByProject) {
    return Array.from(documentsMapByProject.entries()).map(([projectName, documents]) => {
        return {
            projectName,
            duplicates: getDuplicateProjectOperations(documents),
        };
    });
}
function getDuplicateProjectOperations(documents) {
    const operations = new Map();
    Array.from(documents.entries()).forEach(([filePath, document]) => {
        document.definitions.filter(isOperationDefinition).forEach((definition) => {
            const { name } = definition;
            if (name && name.value) {
                const map = operations.get(name.value);
                if (map) {
                    map.add(filePath);
                }
                else {
                    operations.set(name.value, new Set([filePath]));
                }
            }
        });
    });
    return Array.from(operations.entries())
        .filter(([, filePaths]) => filePaths.size > 1)
        .map(([operationName, filePath]) => {
        return { operationName, filePaths: Array.from(filePath) };
    });
}
function isOperationDefinition(definition) {
    return definition.kind === 'OperationDefinition';
}
